@@ -170,6 +170,7 @@ class auto_timeseries:
         self.growth = "linear"
         self.allowed_models = ['best', 'prophet', 'stats', 'ml', 'arima','ARIMA','Prophet','SARIMAX', 'VAR', 'ML']
         self.dask_xgboost_flag = dask_xgboost_flag
+        self.sep = ','
 
         # new function.
         if args:
@@ -223,7 +224,7 @@ class auto_timeseries:
         :type sep Optional[str]
 
         """
-
+        self.sep = sep
         list_of_valid_time_ints = ['B','C','D','W','M','SM','BM','CBM',
                                         'MS','SMS','BMS','CBMS','Q','BQ','QS','BQS',
                                         'A,Y','BA,BY','AS,YS','BAS,BYS','BH',
@@ -284,7 +285,10 @@ class auto_timeseries:
         if isinstance(traindata, str):
             if traindata != '':
                 try:
-                    ts_df = load_ts_data(traindata, self.ts_column, sep, target, self.dask_xgboost_flag)
+                    if self.dask_xgboost_flag:
+                        dask_df, ts_df = load_ts_data(traindata, self.ts_column, sep, target, self.dask_xgboost_flag)
+                    else:
+                        _, ts_df = load_ts_data(traindata, self.ts_column, sep, target, self.dask_xgboost_flag)
                     if isinstance(ts_df, str):
                         print("""Time Series column '%s' could not be converted to a Pandas date time column.
                             Please convert your ts_column into a pandas date-time and try again""" %self.ts_column)
@@ -294,25 +298,21 @@ class auto_timeseries:
                             print('    Dask Dataframe loaded successfully. Shape of data set = (%s,%s)' %(
                                                 ts_df.shape[0].compute(),ts_df.shape[1]))
                         else:
-                            print('    File loaded successfully. Shape of data set = %s' %(ts_df.shape,))
+                            print('    File loaded into pandas dataframe successfully. Shape of data set = %s' %(ts_df.shape,))
                 except Exception:
                     print('File could not be loaded. Check the path or filename and try again')
                     return None
         elif isinstance(traindata, pd.DataFrame):
-            
             print('Input is data frame. Performing Time Series Analysis')
             print(f"ts_column: {self.ts_column} sep: {sep} target: {target}")
-            ts_df = load_ts_data(traindata, self.ts_column, sep, target, self.dask_xgboost_flag)
+            if self.dask_xgboost_flag:
+                dask_df, ts_df = load_ts_data(traindata, self.ts_column, sep, target, self.dask_xgboost_flag)
+            else:
+                _, ts_df = load_ts_data(traindata, self.ts_column, sep, target, self.dask_xgboost_flag)
             if isinstance(ts_df, str):
                 print("""Time Series column '%s' could not be converted to a Pandas date time column.
                     Please convert your input into a date-time column  and try again""" %self.ts_column)
                 return None
-            else:
-                if type(ts_df) == dask.dataframe.core.DataFrame:
-                    print('    Dask Dataframe loaded successfully. Shape of data set = (%s,%s)' %(
-                                        ts_df.shape[0].compute(),ts_df.shape[1]))
-                else:
-                    print('    pandas Dataframe loaded successfully. Shape of data set = %s' %(ts_df.shape,))
         else:
             print('File name is an empty string. Please check your input and try again')
             return None
@@ -664,7 +664,8 @@ class auto_timeseries:
 
         if self.__any_contained_in_list(what_list=['ml', 'ML','best'], in_list=self.model_type):
             ########## Let's build a Machine Learning Model now with Time Series Data ################
-
+            ### Remember that if the dask_xgboost_flag is set to True, we will use Dask DF to build model
+            ###  Otherwise, if the dask_xgboost_flag is set to False, we will use pandas DF to build model
             print("\n")
             print("="*50)
             print("Building ML Model")
@@ -699,16 +700,28 @@ class auto_timeseries:
                 forecast_period = self.forecast_period,
                 ts_column = self.ts_column,
                 time_interval = self.time_interval,
+                sep = self.sep,
+                dask_xgboost_flag = self.dask_xgboost_flag,
                 verbose=self.verbose)
             try:
-                # best = model_build.fit(ts_df=ts_df, target_col=target, lags=lag)
-                model, forecasts, rmse, norm_rmse = model_build.fit(
-                    ts_df=ts_df,
-                    target_col=target,
-                    ts_column = self.ts_column,
-                    cv = cv,
-                    lags=lag
-                )
+                if self.dask_xgboost_flag:
+                    ### We must use dask_df to build a model #####
+                    model, forecasts, rmse, norm_rmse = model_build.fit(
+                        ts_df=dask_df,
+                        target_col=target,
+                        ts_column = self.ts_column,
+                        cv = cv,
+                        lags=lag
+                    )
+                else:
+                    ### We must use pandas ts_df to build a model #####
+                    model, forecasts, rmse, norm_rmse = model_build.fit(
+                        ts_df=ts_df,
+                        target_col=target,
+                        ts_column = self.ts_column,
+                        cv = cv,
+                        lags=lag
+                    )
 
                 if self.score_type == 'rmse':
                     score_val = rmse
@@ -773,19 +786,21 @@ class auto_timeseries:
         f1_stats = {}
         for key, _ in self.ml_dict.items():
             cv_scores = self.ml_dict[key][self.score_type]
-
             # Standardize to a list
-            if len(self.ml_dict[key][self.score_type]) == 0:
+            if self.ml_dict[key][self.score_type] == np.inf:
                 f1_stats[key] = np.inf
             else:
-                if isinstance(cv_scores, np.ndarray):
-                    cv_scores = cv_scores.tolist()
-                if not isinstance(cv_scores, List):
-                    cv_scores = [cv_scores]
-                if len(cv_scores) == 0:
-                        f1_stats[key] = np.inf
+                if len(self.ml_dict[key][self.score_type]) == 0:
+                    f1_stats[key] = np.inf
                 else:
-                    f1_stats[key] = sum(cv_scores)/len(cv_scores)
+                    if isinstance(cv_scores, np.ndarray):
+                        cv_scores = cv_scores.tolist()
+                    elif not isinstance(cv_scores, list):
+                        cv_scores = [cv_scores]
+                    if len(cv_scores) == 0:
+                            f1_stats[key] = np.inf
+                    else:
+                        f1_stats[key] = sum(cv_scores)/len(cv_scores)
 
         best_model_name = min(f1_stats.items(), key=operator.itemgetter(1))[0]
         return best_model_name
@@ -834,11 +849,12 @@ class auto_timeseries:
         model: str = '',
         simple: bool = False,
         time_interval: str = '',
+        saved_model: str = '',
         ):
         """
         Predict the results
         """
-
+        
         if isinstance(testdata, pd.Series) or isinstance(testdata, pd.DataFrame):
             # During training, we internally converted a column datetime index to the dataframe date time index
             # We need to do the same while predicing for consistence
@@ -846,58 +862,85 @@ class auto_timeseries:
                 print('Predicting using test dataframe as input for ML model')
             else:
                 print('Predicting using test dataframe as input for %s model' %self.get_best_model_name())
+        elif type(testdata) == dask.dataframe.core.DataFrame:
+            # During training, we internally converted a column datetime index to the dataframe date time index
+            # We need to do the same while predicing for consistence
+            if (model == 'ML') or self.get_best_model_name() == 'ML' or (model == 'best' and self.get_best_model_name() == 'ML'):
+                ### Now do the predictions using the final model asked to be predicted ###
+                print('Predicting using test dask data frame as input for ML model')
+            else:
+                print('Cannot predicting using dask dataframe as input for %s model. Returning' %self.get_best_model_name())
+                return
         elif isinstance(testdata, int):
             #### if testdata is an Integer, then it appears to be a forecast period, then use it that way
             ### only certain stats-based models can use forecast period
             if (model == 'ML') or (model == 'best' and self.get_best_model_name() == 'ML'):
                 print('Error in input: You must give a test dataframe for ML models')
-            else:
-                
+            else:                
                 print('Predicting using forecast period=%s as input for %s model' %(testdata, self.get_best_model_name()))
 
+        ##### This is where we load the model to predict ###########
         if isinstance(model, str):
             if model == '':
                 bestmodel = self.get_best_model_build()
             elif model.lower() == 'best':
                 bestmodel = self.get_best_model_build()
             else:
+                if saved_model:
+                    ### if a saved_model is specified, just use that model as given and use it ###
+                    print('Make sure the given saved_model has already been trained on this data. Otherwise, it will give errors.')
+                    self.ml_dict[model]['model'] = copy.deepcopy(saved_model)
+                bestmodel = self.get_model_build(model)
                 if self.get_model_build(model) is not None:
                     bestmodel = self.get_model_build(model)
                 else:
-                    print(f"(Error) Model of type '{model}' does not exist. No predictions will be made.")
+                    print(f"(Error) Model of type '{model}' does not exist. Check inputs and try again.")
                     return None
             self.model = bestmodel
         else:
             ### if no model is specified, just use the best model ###
             bestmodel = self.get_best_model_build()
             self.model = bestmodel
-
-
-        if isinstance(testdata, pd.Series) or isinstance(testdata, pd.DataFrame):
-            # During training, we internally converted a column datetime index to the dataframe date time index
-            # We need to do the same while predicing for consistence
-            if (model == 'ML') or self.get_best_model_name() == 'ML' or (model == 'best' and self.get_best_model_name() == 'ML'):
-                ### Now do the predictions using the final model asked to be predicted ###
-                predictions = bestmodel.predict(testdata,simple=simple, time_interval=time_interval)
-            elif model.lower() == 'prophet' or self.get_best_model_name() == 'Prophet':
-                predictions = bestmodel.predict(testdata,simple=simple)
-            elif self.get_best_model_name() == 'VAR':
-                predictions = bestmodel.predict(testdata,simple=simple)
+        ##### Now make predictions using model given ######
+        try:
+            if isinstance(testdata, pd.Series) or isinstance(testdata, pd.DataFrame):
+                # During training, we internally converted a column datetime index to the dataframe date time index
+                # We need to do the same while predicing for consistence
+                if (model == 'ML') or self.get_best_model_name() == 'ML' or (model == 'best' and self.get_best_model_name() == 'ML'):
+                    ### Now do the predictions using the final model asked to be predicted ###
+                    predictions = bestmodel.predict(testdata,simple=simple, time_interval=time_interval)
+                elif model.lower() == 'prophet' or self.get_best_model_name() == 'Prophet':
+                    predictions = bestmodel.predict(testdata,simple=simple)
+                elif self.get_best_model_name() == 'VAR':
+                    predictions = bestmodel.predict(testdata,simple=simple)
+                else:
+                    predictions = bestmodel.predict(testdata,simple=simple)
+            elif type(testdata) == dask.dataframe.core.DataFrame:
+                # During training, we internally converted a column datetime index to the dataframe date time index
+                # We need to do the same while predicing for consistence
+                if (model == 'ML') or self.get_best_model_name() == 'ML' or (model == 'best' and self.get_best_model_name() == 'ML'):
+                    ### Now do the predictions using the final model asked to be predicted ###
+                    predictions = bestmodel.predict(testdata,simple=simple, time_interval=time_interval)
+                else:
+                    print('Dask dataframes are not accepted as input for non-ML models. Please change your input and try again.')
+                    predictions = np.array()
+                    return predictions
+            elif isinstance(testdata, int):
+                #### if testdata is an Integer, then it appears to be a forecast period, then use it that way
+                ### only certain stats-based models can use forecast period
+                if (model == 'ML') or (model == 'best' and self.get_best_model_name() == 'ML'):
+                    print(f'{model} is an ML-based model, hence it cannot be used with a forecast period')
+                    predictions = None
+                else:
+                    predictions = bestmodel.predict(testdata,simple=simple)
             else:
-                predictions = bestmodel.predict(testdata,simple=simple)
-        elif isinstance(testdata, int):
-            #### if testdata is an Integer, then it appears to be a forecast period, then use it that way
-            ### only certain stats-based models can use forecast period
-            if (model == 'ML') or (model == 'best' and self.get_best_model_name() == 'ML'):
-                print(f'{model} is an ML-based model, hence it cannot be used with a forecast period')
-                predictions = None
-            else:
-                predictions = bestmodel.predict(testdata,simple=simple)
-        else:
-            ### if there is no testdata, at least they must give forecast_period
-            if testdata is None:
-                print('If test_data is None, then forecast_period must be given')
-                return
+                ### if there is no testdata, at least they must give forecast_period
+                if testdata is None:
+                    print('If test_data is None, then forecast_period must be given')
+                    return
+        except:
+            print('Model is erroring during prediction. Check inputs and try again.')
+            return
         return predictions
 
     def get_leaderboard(self, ascending=True) -> Optional[pd.DataFrame]:
@@ -997,7 +1040,7 @@ def get_mean_cv_score(cv_scores: Union[float, List]):
 
 #################################################################################
 module_type = 'Running' if  __name__ == "__main__" else 'Imported'
-version_number = '0.0.43'
+version_number = '0.0.45'
 print(f"""{module_type} auto_timeseries version:{version_number}. Call by using:
 model = auto_timeseries(score_type='rmse',
         time_interval='M', non_seasonal_pdq=None, seasonality=False,
